"""qfitter.py -"""

import math

import matplotlib.pyplot as plt
import numpy as np
from scipy.constants import h, physical_constants
from tqdm import tqdm

phi0 = physical_constants["mag. flux quantum"][0]

try:
    import cupy as cp  # type: ignore
except ImportError:
    import numpy as cp


def band(n):
    """band matrix: M(i,j)=kron(i,j-1)"""
    band = np.array(np.zeros((n, n), dtype=np.complex64))
    for i in range(n - 1):
        band[i][i + 1] = 1
    return band


def dest(n):
    """annihilation operator: M(i,j)=sqrt(i)*kron(i,j-1)"""
    dest = np.array(np.zeros((n, n), dtype=np.complex64))
    for i in range(n - 1):
        dest[i][i + 1] = np.sqrt(i + 1)
    return dest


def nsq(n):
    nsq = np.array(np.zeros((2 * n + 1, 2 * n + 1), dtype=np.complex64))
    for i in range(2 * n + 1):
        nsq[i][i] = (i - n) ** 2
    return nsq


def nmat(n):
    nmat = np.array(np.zeros((2 * n + 1, 2 * n + 1), dtype=np.complex64))
    for i in range(2 * n + 1):
        nmat[i][i] = i - n
    return nmat


def phi1_mat(nmax):
    pb = np.array(np.zeros((2 * nmax + 1, 2 * nmax + 1), dtype=np.complex64))
    for m1 in range(-nmax, nmax + 1):
        for m2 in range(-nmax, nmax + 1):
            if m1 != m2:
                pb[m1 + nmax][m2 + nmax] = (1j * (-1) ** (m1 - m2)) / (m1 - m2)
    return pb


def phi2_mat(nmax):
    pb2 = np.array(np.zeros((2 * nmax + 1, 2 * nmax + 1), dtype=np.complex64))
    for m1 in range(-nmax, nmax + 1):
        for m2 in range(-nmax, nmax + 1):
            if m1 != m2:
                pb2[m1 + nmax][m2 + nmax] = (2 * (-1) ** (m1 - m2)) / (m1 - m2) ** 2
            else:
                pb2[m1 + nmax][m2 + nmax] = np.pi**2 / 3
    return pb2


sigz = [[1, 0], [0, -1]]
sigx = [[0, 1], [1, 0]]


def QHami(
    N: int, eps: float, g: float, d: float, omegaC0: float, L: float, R: float
) -> np.ndarray:
    """Get a Hamiltonian of ...

    Args:
        N: the number of dimension for the cavity (size of fock space)
        eps: xxx

    Return:
        Hamiltonian - np.ndarray[complex], dims=[N,N]
    """
    sz = np.kron(sigz, np.eye(N))
    sx = np.kron(sigx, np.eye(N))
    a = np.kron(np.eye(2), dest(N))
    if eps > 0:
        omega = omegaC0 * (1 + R * eps)
    else:
        omega = omegaC0 * (1 - L * eps)
    H = (
        eps * sz / 2
        + d * sx / 2
        + g * np.matmul(sz, a.T + a)
        + omega * np.matmul(a.T, a)
    )
    return H


def Rabi(flist, params):
    """Rabi-model Hamiltonian for fitting"""
    F = 13
    bn = 10
    E1 = [[] for i in range(bn)]
    for f in flist:
        H = QHami(
            F, f * params[5], params[0], params[1], params[2], params[3], params[4]
        )
        evals, ekets = np.linalg.eigh(H)
        for q in range(bn):
            E1[q].append(evals[q])
    return E1


def Rabi05(flist, params, df):
    F = 13
    bn = 10
    E1 = [[] for i in range(bn)]
    for f in flist:
        H = QHami(
            F,
            2 * (f - 0.5) * params[5] * df,
            params[0],
            params[1],
            params[2],
            params[3],
            params[4],
        )
        evals, ekets = np.linalg.eigh(H)
        for q in range(bn):
            E1[q].append(evals[q])
    return E1


def circuit_spectrum(phi_list, params, nlist):
    Ej = params[0]
    Ec = params[1]
    Lr = params[2]
    Cr = params[3]
    alpha = params[4]
    beta = params[5]
    b = 1.0

    wr = 1e-9 / (2 * np.pi * np.sqrt(Cr * Lr))
    Ir = np.sqrt(h / (4 * np.pi * Lr * np.sqrt(Lr * Cr)))

    bn = 10  # haow many band will be shown
    E = [[] for i in range(bn)]
    E1 = [[] for i in range(bn)]
    nmax1 = nlist[0]
    nmax2 = nlist[1]
    nmax3 = nlist[2]
    fock = 15
    dim1 = 2 * nmax1 + 1
    dim2 = 2 * nmax2 + 1
    dim3 = 2 * nmax3 + 1

    # Matrix preparation
    eye1 = np.eye(dim1)
    eye2 = np.eye(dim2)
    eye3 = np.eye(dim3)
    eyef = np.eye(fock)

    # Hamiltonian construction
    # Kinetic energy part
    Nbt = np.kron(np.kron(nsq(nmax1), eye2), eye3)
    Nal = np.kron(np.kron(eye1, nsq(nmax2)), eye3)
    Nb = np.kron(eye1, np.kron(eye2, nsq(nmax3)))
    NbtNal = np.kron(np.kron(nmat(nmax1), nmat(nmax2)), eye3)
    NalNb = np.kron(np.kron(eye1, nmat(nmax2)), nmat(nmax3))
    NbtNb = np.kron(nmat(nmax1), np.kron(eye2, nmat(nmax3)))

    # Resonator
    Hr0 = wr * np.matmul(dest(fock).T, dest(fock))

    # phi and phi**2 term
    Pb = np.kron(phi1_mat(nmax1), np.kron(eye2, eye3))
    Pb2 = np.kron(phi2_mat(nmax1), np.kron(eye2, eye3))
    Hcr = dest(fock).T + dest(fock)
    coef = -Ir * (phi0 / (2 * np.pi)) / (h * 1e9)

    # Flux independent potential
    HL = (0.5 / (Lr * h * 1e9)) * (phi0 / (2 * np.pi)) ** 2 * Pb2

    band1 = np.kron(np.kron(band(dim1) + band(dim1).T, eye2), eye3)
    band2 = np.kron(np.kron(eye1, band(dim2) + band(dim2).T), eye3)
    band3 = np.kron(np.kron(eye1, eye2), band(dim3) + band(dim3).T)
    Hj = -0.5 * Ej * (beta * band1 + alpha * band2 + band3)
    bandf = np.kron(np.kron(band(dim1), band(dim2)), band(dim3))

    D = alpha * b * beta + alpha * b + alpha * beta + b * beta
    Cb = alpha * b + alpha + b
    C1 = b * beta + b + beta
    C2 = alpha * b + alpha * beta + b * beta
    Ht = (4 * Ec / D) * (
        Cb * Nbt
        + C1 * Nal
        + C2 * Nb
        - 2 * b * NbtNal
        - 2 * b * beta * NalNb
        - 2 * alpha * beta * NbtNb
    )

    for phi in tqdm(phi_list):  # Flux dependent part
        Hf = (
            -0.5
            * b
            * Ej
            * (bandf * np.exp(2j * np.pi * phi) + bandf.T * np.exp(-2j * np.pi * phi))
        )

        Hq0 = Ht + HL + Hj + Hf
        evals0, ekets0 = np.linalg.eigh(Hq0)

        for q in range(bn):
            E[q].append(evals0[q])
        qspace = 25
        psiq = []
        for k in range(qspace):
            psiq.append(ekets0[:, k])
        phib_mat = np.array(np.zeros((qspace, qspace), dtype=np.cdouble))
        hq0 = np.array(np.zeros((qspace, qspace), dtype=np.cdouble))
        for x in range(qspace):
            hq0[x][x] = evals0[x]
            for y in range(qspace):
                phib_mat[x][y] = np.dot(np.conjugate(psiq[x]), np.dot(Pb, psiq[y]))

        Hqco = np.kron(hq0, eyef)
        Hcco = coef * np.kron(phib_mat, Hcr)
        Hrco = np.kron(np.eye(qspace), Hr0)
        evals1, ekets1 = np.linalg.eigh(Hqco + Hcco + Hrco)

        for q in range(bn):
            E1[q].append(evals1[q])
    eps0 = np.array(E[1]) - np.array(E[0])
    eps = np.sign(phi_list - 0.5) * np.sqrt(abs(eps0**2 - min(eps0) ** 2))
    return E1, eps


def circuit_spectrum_gpu(phi_list, params, nlist):
    Ej = params[0]
    Ec = params[1]
    Lr = params[2]
    Cr = params[3]
    alpha = params[4]
    beta = params[5]
    b = 1.0

    wr = 1e-9 / (2 * np.pi * np.sqrt(Cr * Lr))
    Ir = np.sqrt(h / (4 * np.pi * Lr * np.sqrt(Lr * Cr)))

    bn = 10  # haow many band will be shown
    E = [[] for i in range(bn)]
    E1 = [[] for i in range(bn)]
    nmax1 = nlist[0]
    nmax2 = nlist[1]
    nmax3 = nlist[2]
    fock = 15
    dim1 = 2 * nmax1 + 1
    dim2 = 2 * nmax2 + 1
    dim3 = 2 * nmax3 + 1

    # Matrix preparation
    eye1 = np.eye(dim1)
    eye2 = np.eye(dim2)
    eye3 = np.eye(dim3)
    eyef = np.eye(fock)

    # Hamiltonian construction
    # Kinetic energy part
    Nbt = np.kron(np.kron(nsq(nmax1), eye2), eye3)
    Nal = np.kron(np.kron(eye1, nsq(nmax2)), eye3)
    Nb = np.kron(eye1, np.kron(eye2, nsq(nmax3)))
    NbtNal = np.kron(np.kron(nmat(nmax1), nmat(nmax2)), eye3)
    NalNb = np.kron(np.kron(eye1, nmat(nmax2)), nmat(nmax3))
    NbtNb = np.kron(nmat(nmax1), np.kron(eye2, nmat(nmax3)))

    # Resonator
    Hr0 = wr * np.matmul(dest(fock).T, dest(fock))

    # phi and phi**2 term
    Pb = np.kron(phi1_mat(nmax1), np.kron(eye2, eye3))
    Pb2 = np.kron(phi2_mat(nmax1), np.kron(eye2, eye3))
    Hcr = dest(fock).T + dest(fock)
    coef = -Ir * (phi0 / (2 * np.pi)) / (h * 1e9)

    # Flux independent potential
    HL = (0.5 / (Lr * h * 1e9)) * (phi0 / (2 * np.pi)) ** 2 * Pb2

    band1 = np.kron(np.kron(band(dim1) + band(dim1).T, eye2), eye3)
    band2 = np.kron(np.kron(eye1, band(dim2) + band(dim2).T), eye3)
    band3 = np.kron(np.kron(eye1, eye2), band(dim3) + band(dim3).T)
    Hj = -0.5 * Ej * (beta * band1 + alpha * band2 + band3)
    bandf = np.kron(np.kron(band(dim1), band(dim2)), band(dim3))

    D = alpha * b * beta + alpha * b + alpha * beta + b * beta
    Cb = alpha * b + alpha + b
    C1 = b * beta + b + beta
    C2 = alpha * b + alpha * beta + b * beta
    Ht = (4 * Ec / D) * (
        Cb * Nbt
        + C1 * Nal
        + C2 * Nb
        - 2 * b * NbtNal
        - 2 * b * beta * NalNb
        - 2 * alpha * beta * NbtNb
    )

    for phi in tqdm(phi_list):  # Flux dependent part
        Hf = (
            -0.5
            * b
            * Ej
            * (bandf * np.exp(2j * np.pi * phi) + bandf.T * np.exp(-2j * np.pi * phi))
        )

        Hq0 = cp.array(Ht + HL + Hj + Hf, dtype=cp.complex64)
        evals01, ekets01 = cp.linalg.eigh(Hq0)
        evals0 = evals01.get()
        ekets0 = ekets01.get()

        for q in range(bn):
            E[q].append(evals0[q])
        qspace = 25
        psiq = []
        for k in range(qspace):
            psiq.append(ekets0[:, k])
        phib_mat = np.array(np.zeros((qspace, qspace), dtype=np.cdouble))
        hq0 = np.array(np.zeros((qspace, qspace), dtype=np.cdouble))
        for x in range(qspace):
            hq0[x][x] = evals0[x]
            for y in range(qspace):
                phib_mat[x][y] = np.dot(np.conjugate(psiq[x]), np.dot(Pb, psiq[y]))

        Hqco = np.kron(hq0, eyef)
        Hcco = coef * np.kron(phib_mat, Hcr)
        Hrco = np.kron(np.eye(qspace), Hr0)
        H_final = cp.array(Hqco + Hcco + Hrco, dtype=cp.complex64)
        evals11, ekets11 = cp.linalg.eigh(H_final)
        evals1 = evals11.get()
        ekets1 = ekets11.get()

        for q in range(bn):
            E1[q].append(evals1[q])
    eps0 = np.array(E[1]) - np.array(E[0])
    eps = np.sign(phi_list - 0.5) * np.sqrt(abs(eps0**2 - min(eps0) ** 2))
    return E1, eps


def circuit_spectrum_Q(phi_list, params, nlist):
    Ej = params[0]
    Ec = params[1]
    Lr = params[2]
    Cr = params[3]
    alpha = params[4]
    beta = params[5]
    b = 1.0

    wr = 1e-9 / (2 * np.pi * np.sqrt(Cr * Lr))
    Ir = np.sqrt(h / (4 * np.pi * Lr * np.sqrt(Lr * Cr)))

    bn = 10  # haow many band will be shown
    E = [[] for i in range(bn)]
    E1 = [[] for i in range(bn)]
    nmax1 = nlist[0]
    nmax2 = nlist[1]
    nmax3 = nlist[2]
    fock = 15
    dim1 = 2 * nmax1 + 1
    dim2 = 2 * nmax2 + 1
    dim3 = 2 * nmax3 + 1

    # Matrix preparation
    eye1 = np.eye(dim1)
    eye2 = np.eye(dim2)
    eye3 = np.eye(dim3)
    eyef = np.eye(fock)

    # Hamiltonian construction
    # Kinetic energy part
    Nbt = np.kron(np.kron(nsq(nmax1), eye2), eye3)
    Nal = np.kron(np.kron(eye1, nsq(nmax2)), eye3)
    Nb = np.kron(eye1, np.kron(eye2, nsq(nmax3)))
    NbtNal = np.kron(np.kron(nmat(nmax1), nmat(nmax2)), eye3)
    NalNb = np.kron(np.kron(eye1, nmat(nmax2)), nmat(nmax3))
    NbtNb = np.kron(nmat(nmax1), np.kron(eye2, nmat(nmax3)))

    # Resonator
    Hr0 = wr * np.matmul(dest(fock).T, dest(fock))

    # phi and phi**2 term
    Pb = np.kron(phi1_mat(nmax1), np.kron(eye2, eye3))
    Pb2 = np.kron(phi2_mat(nmax1), np.kron(eye2, eye3))
    Hcr = dest(fock).T + dest(fock)
    coef = -Ir * (phi0 / (2 * np.pi)) / (h * 1e9)

    # Flux independent potential
    HL = (0.5 / (Lr * h * 1e9)) * (phi0 / (2 * np.pi)) ** 2 * Pb2

    band1 = np.kron(np.kron(band(dim1) + band(dim1).T, eye2), eye3)
    band2 = np.kron(np.kron(eye1, band(dim2) + band(dim2).T), eye3)
    band3 = np.kron(np.kron(eye1, eye2), band(dim3) + band(dim3).T)
    Hj = -0.5 * Ej * (beta * band1 + alpha * band2 + band3)
    bandf = np.kron(np.kron(band(dim1), band(dim2)), band(dim3))

    D = alpha * b * beta + alpha * b + alpha * beta + b * beta
    Cb = alpha * b + alpha + b
    C1 = b * beta + b + beta
    C2 = alpha * b + alpha * beta + b * beta
    Ht = (4 * Ec / D) * (
        Cb * Nbt
        + C1 * Nal
        + C2 * Nb
        - 2 * b * NbtNal
        - 2 * b * beta * NalNb
        - 2 * alpha * beta * NbtNb
    )

    for phi in tqdm(phi_list):  # Flux dependent part
        Hf = (
            -0.5
            * b
            * Ej
            * (bandf * np.exp(2j * np.pi * phi) + bandf.T * np.exp(-2j * np.pi * phi))
        )

        Hq0 = Ht + HL + Hj + Hf
        evals0, ekets0 = np.linalg.eigh(Hq0)

        for q in range(bn):
            E1[q].append(evals0[q])

    return np.array(E1)


def circuit_spectrum_Qket(phi_list, params, nlist):
    Ej = params[0]
    Ec = params[1]
    Lr = params[2]
    Cr = params[3]
    alpha = params[4]
    beta = params[5]
    b = 1.0

    nmax1 = nlist[0]
    nmax2 = nlist[1]
    nmax3 = nlist[2]
    dim1 = 2 * nmax1 + 1
    dim2 = 2 * nmax2 + 1
    dim3 = 2 * nmax3 + 1

    # Matrix preparation
    eye1 = np.eye(dim1)
    eye2 = np.eye(dim2)
    eye3 = np.eye(dim3)

    # Hamiltonian construction
    # Kinetic energy part
    Nbt = np.kron(np.kron(nsq(nmax1), eye2), eye3)
    Nal = np.kron(np.kron(eye1, nsq(nmax2)), eye3)
    Nb = np.kron(eye1, np.kron(eye2, nsq(nmax3)))
    NbtNal = np.kron(np.kron(nmat(nmax1), nmat(nmax2)), eye3)
    NalNb = np.kron(np.kron(eye1, nmat(nmax2)), nmat(nmax3))
    NbtNb = np.kron(nmat(nmax1), np.kron(eye2, nmat(nmax3)))

    # phi**2 term
    Pb2 = np.kron(phi2_mat(nmax1), np.kron(eye2, eye3))

    # Flux independent potential
    HL = (0.5 / (Lr * h * 1e9)) * (phi0 / (2 * np.pi)) ** 2 * Pb2

    band1 = np.kron(np.kron(band(dim1) + band(dim1).T, eye2), eye3)
    band2 = np.kron(np.kron(eye1, band(dim2) + band(dim2).T), eye3)
    band3 = np.kron(np.kron(eye1, eye2), band(dim3) + band(dim3).T)
    Hj = -0.5 * Ej * (beta * band1 + alpha * band2 + band3)
    bandf = np.kron(np.kron(band(dim1), band(dim2)), band(dim3))

    D = alpha * b * beta + alpha * b + alpha * beta + b * beta
    Cb = alpha * b + alpha + b
    C1 = b * beta + b + beta
    C2 = alpha * b + alpha * beta + b * beta
    Ht = (4 * Ec / D) * (
        Cb * Nbt
        + C1 * Nal
        + C2 * Nb
        - 2 * b * NbtNal
        - 2 * b * beta * NalNb
        - 2 * alpha * beta * NbtNb
    )
    Ev = []
    Ek = []
    for phi in tqdm(phi_list):  # Flux dependent part
        Hf = (
            -0.5
            * b
            * Ej
            * (bandf * np.exp(2j * np.pi * phi) + bandf.T * np.exp(-2j * np.pi * phi))
        )
        Hq0 = Ht + HL + Hj + Hf
        evals0, ekets0 = np.linalg.eigh(Hq0)
        Ev.append(evals0)
        Ek.append(ekets0)
    return Ev, Ek


def circuit_spectrum_QR(Ev, Ek, params, nlist, qspace):
    Ej = params[0]
    Ec = params[1]
    Lr = params[2]
    Cr = params[3]
    alpha = params[4]
    beta = params[5]
    b = 1.0

    wr = 1e-9 / (2 * np.pi * np.sqrt(Cr * Lr))
    Ir = np.sqrt(h / (4 * np.pi * Lr * np.sqrt(Lr * Cr)))

    bn = 10
    E1 = [[] for i in range(bn)]
    nmax1 = nlist[0]
    nmax2 = nlist[1]
    nmax3 = nlist[2]
    fock = 15
    dim1 = 2 * nmax1 + 1
    dim2 = 2 * nmax2 + 1
    dim3 = 2 * nmax3 + 1

    # Matrix preparation
    eye1 = np.eye(dim1)
    eye2 = np.eye(dim2)
    eye3 = np.eye(dim3)
    eyef = np.eye(fock)

    # Resonator
    Hr0 = wr * np.matmul(dest(fock).T, dest(fock))
    Pb = np.kron(phi1_mat(nmax1), np.kron(eye2, eye3))
    Hcr = dest(fock).T + dest(fock)
    coef = -Ir * (phi0 / (2 * np.pi)) / (h * 1e9)

    for i in tqdm(range(len(Ev))):
        evals0 = Ev[i]
        ekets0 = Ek[i]
        psiq = []
        for k in range(qspace):
            psiq.append(ekets0[:, k])
        phib_mat = np.array(np.zeros((qspace, qspace), dtype=np.cdouble))
        hq0 = np.array(np.zeros((qspace, qspace), dtype=np.cdouble))
        for x in range(qspace):
            hq0[x][x] = evals0[x]
            for y in range(qspace):
                phib_mat[x][y] = np.dot(np.conjugate(psiq[x]), np.dot(Pb, psiq[y]))

        Hqco = np.kron(hq0, eyef)
        Hcco = coef * np.kron(phib_mat, Hcr)
        Hrco = np.kron(np.eye(qspace), Hr0)
        evals1, ekets1 = np.linalg.eigh(Hqco + Hcco + Hrco)

        for q in range(bn):
            E1[q].append(evals1[q])
    return E1



"""A module for post-processing to extract peak/dip structures from two-dimensional spectrum."""

import itertools
from math import isclose, isnan
from typing import Callable, Iterable, Mapping, Sequence

import cv2
import matplotlib.pyplot as plt
import numpy as np
from cv2.typing import MatLike
from numpy.typing import ArrayLike, NDArray
from skimage import filters, measure


def normalize(x) -> NDArray[np.floating]:
    """x in R --> [0..1]"""
    x = np.array(x)
    mn = x.min(axis=None, keepdims=True)
    mx = x.max(axis=None, keepdims=True)
    return (x - mn) / (mx - mn)


DEFAULT_KWARGS_SAVEFIG = dict(bbox_inches="tight", pad_inches=0.5, dpi=500)


def apply_image_filter(
    img: MatLike,
    func: Callable[[MatLike], MatLike],
    with_plot=False,
    filename_plot: str | None = None,
    show=True,
    **kwargs,
) -> MatLike:
    """Apply image processing for an arbitrary filter function.

    Args:
        img: (M, N[, P]) ndarray, An input image(color image of 2d spectrum)
        func: callable, A function implementing an image processing filter. The first argument is `img`.
        with_plot: if True, make a matplotlib plot.
        filename_plot: ...
        kwargs: will be passed to the function.

    Return:
        (M, N[, P]) ndarray
    """
    result = func(img, **kwargs)

    if with_plot:
        fig, ax = plt.subplots(figsize=(8, 6))
        mappable = ax.pcolor(result, cmap="bwr")
        cbar_num_format = "%.2f"
        plt.colorbar(mappable, ax=ax, format=cbar_num_format)
        plt.tight_layout()
        if filename_plot is not None:
            plt.savefig(filename_plot, **DEFAULT_KWARGS_SAVEFIG)
        if show:
            plt.show()

    return result


def apply_sato_filter(
    img: MatLike,
    sigmas: float | Iterable[float] | NDArray[np.floating] = 4.0,
    black_ridges: bool = False,
    with_plot=False,
    filename_plot: str | None = None,
    show=True,
    **kwargs,
) -> MatLike:
    """Apply image processing (sato-function).

    Args:
        img: (M, N[, P]) ndarray, an input image(color image of 2d spectrum)
        sigmas: TBD
        black_ridges: see <skimage.filters.sato>
        with_plot: if True, make a matplotlib plot.
        filename_plot: ...

    Return:
        (M, N[, P]) ndarray
    """
    img = normalize(img)

    if isinstance(sigmas, (int, float)):
        sigmas = [sigmas]
    kwargs["sigmas"] = sigmas
    kwargs["black_ridges"] = black_ridges

    return apply_image_filter(
        img,
        func=filters.sato,
        with_plot=with_plot,
        filename_plot=filename_plot,
        show=show,
        **kwargs,
    )


def apply_nofilter(
    img: MatLike,
    with_plot=False,
    filename_plot: str | None = None,
    show=True,
    **kwargs,
) -> MatLike:
    """Apply image processing without any processing."""
    img = normalize(img)

    return apply_image_filter(
        img,
        func=lambda img: img,
        with_plot=with_plot,
        filename_plot=filename_plot,
        show=show,
        **kwargs,
    )


def _calc_poly_length(polygon: NDArray) -> float:
    """ポリゴンの長さを計算"""
    xs, ys = np.array(polygon).T
    dx = xs[1:] - xs[:-1]
    dy = ys[1:] - ys[:-1]
    ls = dx**2 + dy**2
    return np.sum(ls)


def find_contours(
    img: MatLike,
    level: float | None = None,
    threshold_length: float | None = None,
    with_plot=False,
    filename_plot: str | None = None,
    show=True,
    **kwargs,
) -> list[NDArray[np.floating]]:
    """Find contours from a 2d spectrum.

    1. apply `skimage.measure.find_contours`
    2. pick up some contours that have a certain length (`>= threshold_length`)
    3. (make plot if necessary)

    Args:
        img: (M, N[, P]) ndarray
        level: see <find_contours>
        threshold_length: プロットに表示する閾値
        with_plot: ...
        filename_plot: ...
        kwargs: the other arguments will be passed to find_contours.

    Return:
        list of polygons
    """

    contours = [(poly) for poly in measure.find_contours(img, level, **kwargs)]
    len_list = [_calc_poly_length(poly) for poly in contours]

    if threshold_length is None:
        threshold_length = float(np.median(len_list))
    flg_list = [L >= threshold_length for L in len_list]

    if with_plot:
        fig, axes = plt.subplots(2, 2, figsize=(16, 12))
        (ax, ax_L), (ax_cont, _) = axes

        ax.pcolor(img, cmap="bwr")
        for idx, (poly, flg) in enumerate(zip(contours, flg_list)):
            if not flg:
                continue

            ax.plot(poly[:, 1], poly[:, 0], "k-", linewidth=1)
            ax.plot(poly[:, 1], poly[:, 0], f"C{idx % 10}--", linewidth=2)

            ax_cont.plot(poly[:, 1], poly[:, 0], f"C{idx % 10}-", linewidth=1)
            ax_cont.annotate(
                idx,
                xy=(np.average(poly[:, 1]), np.average(poly[:, 0])),
                fontsize=10,
                color="k",
                bbox={"facecolor": f"C{idx % 10}", "edgecolor": "w", "alpha": 0.3},
            )
        for poly, flg in zip(contours, flg_list):
            if not flg:
                ax.plot(poly[:, 1], poly[:, 0], "k-", linewidth=1, alpha=0.8)

        ax_L.plot(len_list, ds="steps-mid")
        ax_L.axhline(y=threshold_length, color="k", ls="dashed")
        ax_L.grid()
        ax_L.set_ylabel("Contour length")
        ax_L.set_yscale("log")
        ax_L.set_xlabel("# of contours")

        if filename_plot is not None:
            plt.savefig(filename_plot, **DEFAULT_KWARGS_SAVEFIG)
        if show:
            plt.show()

    return contours


def assign_contours(
    cont_list: Sequence[NDArray[np.floating]],
    dict_indexes: dict[str, tuple[int, ...] | None] | None = None,
    with_plot=False,
    filename_plot: str | None = None,
    show=True,
) -> dict[str, list[NDArray[np.floating]]]:
    """Make a dict of contours. key=label, val=contour

    Example:
        >>> assign_contours(cont_list,
        >>>                 {"a": (2, 3), "b": (7, 6), "c": (4, 5)})
        >>> {"a": ..., "b": ..., "c": ...}
    """

    if dict_indexes is None:
        dict_indexes = {"all": None}

    def _extract(idexes: Sequence[int] | None):
        if idexes is None:
            return [v for v in cont_list]
        else:
            return [cont_list[i] for i in idexes]

    cont_dict = {k: _extract(ids) for k, ids in dict_indexes.items()}

    if with_plot:
        fig, ax = plt.subplots(figsize=(8, 6))

        for i, (kw, contours) in enumerate(cont_dict.items()):
            for j, poly in enumerate(contours):
                ax.plot(
                    poly[:, 1],
                    poly[:, 0],
                    f"C{i % 10}-",
                    label=(kw if j == 0 else None),  # 各ポリゴン群で一つだけラベルする
                )
            ax.legend()

        if filename_plot is not None:
            plt.savefig(filename_plot, **DEFAULT_KWARGS_SAVEFIG)
        if show:
            plt.show()

    return cont_dict


def _is_closed(polygon: NDArray[np.floating]) -> bool:
    return len(polygon) > 1 and all(np.isclose(polygon[0], polygon[-1]))


def batched(iterable, n, *, strict=False):
    # batched('ABCDEFG', 3) → ABC DEF G
    if n < 1:
        raise ValueError("n must be at least one")
    iterator = iter(iterable)
    while batch := tuple(itertools.islice(iterator, n)):
        if strict and len(batch) != n:
            raise ValueError("batched(): incomplete batch")
        yield batch


def _make_enclosure(
    polygons: Sequence[NDArray[np.floating]], h: int, w: int
) -> list[NDArray[np.floating]]:
    """ポリゴンのリストを受け取って、閉ポリゴンになるように変更したリストを返す

    仮定
    - コーナーは必ず範囲外にある
    - 包含関係にはならない

    [ [(y,x), (y,x), ...],
      [(y,x), (y,x), ...],
      ...
    ]
    """
    # Note: +yが上辺. y=0, ..., h-1 / x=0, ..., w-1

    # corner case: polygonがエッジ上から始まって同じ点で終わる（未対応）

    # case: multi polygons are found.
    closed_polygons = [p for p in polygons if _is_closed(p)]
    polygons = [p for p in polygons if not _is_closed(p)]
    if len(polygons) == 0:
        return closed_polygons

    # 左回りにエッジを走査して、端点のリストを順に作っていく
    # その後、2個ずつリストに加える。角は閉包に含まれないと仮定しているので、
    # (終点,始点) のペアが順に得られる
    _xval = lambda v: v[1]  # noqa: E731
    _yval = lambda v: v[0]  # noqa: E731
    edge_points = list(itertools.chain.from_iterable([(p[0], p[-1]) for p in polygons]))
    sorted_edge_points = (
        sorted(
            [pos for pos in edge_points if isclose(pos[1], 0.0)],  # left edge
            key=_yval,
            reverse=True,
        )
        + sorted(
            [pos for pos in edge_points if isclose(pos[0], 0.0)],  # bottom
            key=_xval,
        )
        + sorted(
            [pos for pos in edge_points if isclose(pos[1], w - 1)],  # right
            key=_yval,
        )
        + sorted(
            [pos for pos in edge_points if isclose(pos[0], h - 1)],  # top
            key=_xval,
            reverse=True,
        )
    )
    edge_polygons = [
        np.array(pair) for pair in batched(sorted_edge_points, 2, strict=True)
    ]
    polygons = edge_polygons + list(polygons)

    # 順に接続するように並び替え & マージ
    _polygons = [polygons.pop(0)]
    while len(polygons) > 0:
        p_last = _polygons[-1]

        # 端点がpivotに最も近いものを選択
        dist_to_head = [np.linalg.norm(p[0] - p_last[-1]) for p in polygons]
        dist_to_tail = [np.linalg.norm(p[-1] - p_last[-1]) for p in polygons]
        dist_to = np.min(np.column_stack([dist_to_head, dist_to_tail]), axis=1)
        _idx = np.argmin(dist_to)

        p_next = polygons.pop(_idx)
        if isclose(dist_to[_idx], 0):
            if dist_to_head[_idx] > dist_to_tail[_idx]:
                p_next = np.flip(p_next, axis=0)

            _polygons[-1] = np.concatenate([p_last, p_next[1:]])

        else:
            _polygons.append(p_next)
    polygons = list(_polygons)

    return closed_polygons + polygons


def determine_regions(
    img: MatLike,
    cont_dict: Mapping[str, Sequence[NDArray[np.floating]]],
    kernel: int | NDArray = 3,
    with_plot=False,
    filename_plot: str | None = None,
    show=True,
) -> dict[str, MatLike]:
    """Extract regions to freq-determination.

    1. Dilate contours to get band-like regions.
    2. XOR operation to ensure that each band has no overlap.

    Args:
        img:
        cont_dict: key=label, value=list of polygons
        offset: size of dilation
        ...

    Return:
        A dictionary, key=label, value=list of binary images.
    """

    def make_filled_img(contours: Sequence[NDArray[np.floating]]) -> MatLike:
        _img: MatLike = np.zeros(img.shape, np.uint8)

        for poly in _make_enclosure(contours, *img.shape):
            pts = np.flip(poly, axis=1)
            pts = np.array(pts, dtype=np.int32).reshape((-1, 1, 2))
            _img = cv2.fillPoly(_img, [pts], (1,))
        return _img

    # make regeion (binary image)
    region_img_dict = {
        k: make_filled_img(contours) for k, contours in cont_dict.items()
    }

    # dilate regions
    if isinstance(kernel, int):
        kernel = np.ones((kernel, kernel), np.uint8)
    region_dict = {k: cv2.dilate(r, kernel) for k, r in region_img_dict.items()}

    # logical op.
    # A, B, C, ... ==> A, B & not A, C & not A & not B, ...
    # 始めに来た領域を優先して、領域ごとの被りをなくす
    Rt: MatLike = np.zeros(img.shape, np.uint8)
    for k in region_dict.keys():
        # Note: keysではなくregion.itemsにすると終わらなくなる。これは辞書内を直接編集しているため。
        region_dict[k] = cv2.bitwise_and(region_dict[k], cv2.bitwise_not(Rt))
        Rt = cv2.bitwise_or(Rt, region_dict[k])

    if with_plot:
        fig, ax = plt.subplots(figsize=(8, 6))

        for k, r in region_dict.items():
            ax.pcolor(r, alpha=0.2)

        for i, (kw, contours) in enumerate(cont_dict.items()):
            for j, poly in enumerate(contours):
                ax.plot(
                    poly[:, 1],
                    poly[:, 0],
                    f"C{i % 10}-",
                    label=(
                        kw if j == 0 else None
                    ),  # 各閉ポリゴン群で一つだけラベルする
                )

        if filename_plot is not None:
            plt.savefig(filename_plot, **DEFAULT_KWARGS_SAVEFIG)
        if show:
            plt.show()

    return region_dict


def _argmax(xs, ys) -> int:
    """ピーク位置決定"""
    return int(np.nanargmax(ys))


def determine_peak_positions(
    img: MatLike,
    region_dict: Mapping[str, MatLike],
    xaxis: NDArray[np.floating] | Sequence | None = None,
    yaxis: NDArray[np.floating] | Sequence | None = None,
    method: Callable[[NDArray[np.floating], NDArray[np.floating]], int] = _argmax,
    eliminate_nan=True,
    with_plot=False,
    filename_plot: str | None = None,
    show=True,
) -> dict[str, list[tuple[float, float]]]:
    """! TODO あとで書く

    Args:
        img (np.ndarray): _description_
        region_dict (Mapping[str, np.ndarray[int]]): _description_
        xaxis (np.ndarray | Sequence | None, optional): _description_. Defaults to None.
        yaxis (np.ndarray | Sequence | None, optional): _description_. Defaults to None.
        method: ピーク位置を決定するアルゴリズム
        with_plot (bool, optional): _description_. Defaults to False.
        filename_plot (str, optional): _description_. Defaults to None.

    Returns:
        dict[str, list[tuple[float, float]]]: _description_
    """
    assert len(img.shape) == 2
    h, w = img.shape

    if xaxis is None:
        xaxis = range(w)
    xaxis = np.array(xaxis)
    if yaxis is None:
        yaxis = range(h)
    yaxis = np.array(yaxis)
    assert len(xaxis) == img.shape[1]
    assert len(yaxis) == img.shape[0]

    peak_dict: dict[str, list[tuple[float, float]]] = {
        k: [] for k in region_dict.keys()
    }
    for k, region in region_dict.items():
        _img = np.array(img, dtype=np.float128)
        _img[np.where(region == 0)] = np.nan

        for idx_x in range(w):
            try:
                idx_ypeak = method(xaxis, _img[:, idx_x])

            except ValueError:  # all-nan slice, その軸に値が見つからないケース
                peak_dict[k].append((xaxis[idx_x], np.nan))

            else:
                peak_dict[k].append((xaxis[idx_x], yaxis[idx_ypeak]))

    if eliminate_nan:
        peak_dict = {
            k: [v for v in indexes if not isnan(v[1])]
            for k, indexes in peak_dict.items()
        }

    if with_plot:
        fig, (ax, ax_orig, ax_pos) = plt.subplots(1, 3, figsize=(15, 4))
        X, Y = np.meshgrid(xaxis, yaxis)
        ax.pcolor(X, Y, img)
        ax_orig.pcolor(X, Y, img)
        for i, (kw, positions) in enumerate(peak_dict.items()):
            pos = np.array(positions)
            ax.scatter(pos[:, 0], pos[:, 1], c="red", label=kw, marker="x", alpha=0.5)
            ax_pos.scatter(pos[:, 0], pos[:, 1], c=f"C{i % 10}", label=kw, marker=".")
        ax_pos.set_xlim(ax.get_xlim())
        ax_pos.set_ylim(ax.get_ylim())
        ax_pos.legend()
        if filename_plot is not None:
            plt.savefig(filename_plot, **DEFAULT_KWARGS_SAVEFIG)
        if show:
            plt.show()

    return peak_dict



from math import pi

import numpy as np
import pytest

from scfitpy.image_processing import (
    _make_enclosure,
    apply_nofilter,
    apply_sato_filter,
    assign_contours,
    determine_peak_positions,
    determine_regions,
    find_contours,
)


def test_all_singleline():
    # preparation
    prefix = "./outs/test_all_singleline_"
    np.random.seed(42)

    xs = np.linspace(-1, 1, 101)  # sweep-axis, like dc bias
    fs = np.linspace(100, 200, 201)  # frequency-axis
    pts = np.full_like(xs, xs * 10 + np.average(fs))  # dip point: y=y0 (constant)
    rtol = 0.01  # relative tolerance for analysis

    spec = np.array(
        [2 / (0.05**2 + (fs - f0) ** 2) for f0 in pts]
    ).T  # 形状固定のLorentzian
    spec += np.random.normal(size=spec.shape)
    mag = 20 * np.log10(np.abs(spec))

    # run
    apply_nofilter(mag, with_plot=True, filename_plot=prefix + "mag.png")
    mag_filtered = apply_sato_filter(
        mag,
        4,
        black_ridges=False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_filtered.png",
    )
    cont_list = find_contours(
        mag_filtered,
        level=0.06,
        threshold_length=1,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_find_contours.png",
    )
    cont_dict = assign_contours(
        cont_list,
        {"a": (0, 1)},
        with_plot=True,
        show=False,
        filename_plot=prefix + "contours.png",
    )
    region_dict = determine_regions(
        mag,
        cont_dict,
        kernel=5,
        with_plot=True,
        show=False,
        filename_plot=prefix + "regions.png",
    )
    result = determine_peak_positions(
        mag,
        region_dict,
        xaxis=xs,
        yaxis=fs,
        with_plot=True,
        show=False,
        filename_plot=prefix + "peaks.png",
    )

    # check
    assert pytest.approx(xs) == [v for v, _ in result["a"]]
    assert pytest.approx(pts, rtol) == [v for _, v in result["a"]]


def test_all_twoline():
    # preparation
    prefix = "./outs/test_all_twoline_"
    np.random.seed(42)

    xs = np.linspace(-1, 1, 101)  # sweep-axis, like dc bias
    fs = np.linspace(100, 200, 201)  # frequency-axis
    pts1 = np.full_like(xs, xs * 10 + np.average(fs) + 20)  # dip point: y=y0 (constant)
    pts2 = np.full_like(xs, xs * -3 + np.average(fs) - 20)  # dip point: y=y0 (constant)
    rtol = 0.01  # relative tolerance for analysis

    spec = (
        np.array([117 / (0.05**2 + (fs - f0) ** 2) for f0 in pts1]).T
        + np.array([60 / (0.1**2 + (fs - f0) ** 2) for f0 in pts2]).T
    )
    spec += np.random.normal(size=spec.shape)
    mag = 20 * np.log10(np.abs(spec))

    # run
    apply_nofilter(mag, with_plot=True, filename_plot=prefix + "mag.png")
    mag_filtered = apply_sato_filter(
        mag,
        2,
        black_ridges=False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_filtered.png",
    )
    cont_list = find_contours(
        mag_filtered,
        level=0.02,
        threshold_length=2,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_find_contours.png",
    )
    cont_dict = assign_contours(
        cont_list,
        {"upper": (15, 14), "lower": (8, 9)},
        with_plot=True,
        show=False,
        filename_plot=prefix + "contours.png",
    )
    region_dict = determine_regions(
        mag,
        cont_dict,
        kernel=5,
        with_plot=True,
        show=False,
        filename_plot=prefix + "regions.png",
    )
    result = determine_peak_positions(
        mag,
        region_dict,
        xaxis=xs,
        yaxis=fs,
        with_plot=True,
        show=False,
        filename_plot=prefix + "peaks.png",
    )

    # check
    assert pytest.approx(xs) == [v for v, _ in result["upper"]]
    assert pytest.approx(xs) == [v for v, _ in result["lower"]]
    assert pytest.approx(pts1, rtol) == [v for _, v in result["upper"]]
    assert pytest.approx(pts2, rtol) == [v for _, v in result["lower"]]


def test_all_single_curve():
    # preparation
    prefix = "./outs/test_all_single_curve_"
    np.random.seed(42)

    xs = np.linspace(-1, 1, 101)  # sweep-axis, like dc bias
    fs = np.linspace(100, 200, 201)  # frequency-axis
    pts = np.full_like(
        xs, 40 * xs**2 + np.average(fs) - 30
    )  # dip point: y=y0 (constant)
    rtol = 0.01  # relative tolerance for analysis

    amps = 20 * np.ones_like(xs)
    spec = np.array(
        [a / (0.05**2 + (fs - f0) ** 2) for a, f0 in zip(amps, pts)]
    ).T  # 形状固定のLorentzian
    spec += np.random.normal(size=spec.shape)
    mag = 20 * np.log10(np.abs(spec))

    # run
    apply_nofilter(mag, with_plot=True, filename_plot=prefix + "mag.png")
    mag_filtered = apply_sato_filter(
        mag,
        1,
        False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_filtered.png",
    )
    cont_list = find_contours(
        mag_filtered,
        level=0.025,
        threshold_length=20,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_find_contours.png",
    )
    cont_dict = assign_contours(
        cont_list,
        {"a": (156, 162)},
        with_plot=True,
        show=False,
        filename_plot=prefix + "contours.png",
    )
    region_dict = determine_regions(
        mag,
        cont_dict,
        kernel=5,
        with_plot=True,
        show=False,
        filename_plot=prefix + "regions.png",
    )
    result = determine_peak_positions(
        mag,
        region_dict,
        xaxis=xs,
        yaxis=fs,
        eliminate_nan=False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "peaks.png",
    )

    # check
    assert pytest.approx(xs) == [v for v, _ in result["a"]]
    assert pytest.approx(pts, rtol) == [v for _, v in result["a"]]


def test_all_steep_curve():
    # preparation
    prefix = "./outs/test_all_steep_curve_"
    np.random.seed(42)

    xs = np.linspace(-1, 1, 101)  # sweep-axis, like dc bias
    fs = np.linspace(100, 200, 201)  # frequency-axis
    pts = np.full_like(
        xs,
        200 * xs**3 + np.average(fs),  # 左右ではなく上下に突き抜ける場合
    )  # dip point: y=y0 (constant)
    rtol = 0.01  # relative tolerance for analysis

    amps = 20 * np.ones_like(xs)
    spec = np.array(
        [a / (0.05**2 + (fs - f0) ** 2) for a, f0 in zip(amps, pts)]
    ).T  # 形状固定のLorentzian
    spec += np.random.normal(size=spec.shape)
    mag = 20 * np.log10(np.abs(spec))

    # run
    apply_nofilter(mag, with_plot=True, filename_plot=prefix + "mag.png")
    mag_filtered = apply_sato_filter(
        mag,
        1,
        False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_filtered.png",
    )
    cont_list = find_contours(
        mag_filtered,
        level=0.025,
        threshold_length=1000,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_find_contours.png",
    )
    cont_dict = assign_contours(
        cont_list,
        {"a": (1, 2)},
        with_plot=True,
        show=False,
        filename_plot=prefix + "contours.png",
    )
    region_dict = determine_regions(
        mag,
        cont_dict,
        kernel=5,
        with_plot=True,
        show=False,
        filename_plot=prefix + "regions.png",
    )
    result = determine_peak_positions(
        mag,
        region_dict,
        xaxis=xs,
        yaxis=fs,
        eliminate_nan=False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "peaks.png",
    )

    # check
    assert pytest.approx(xs) == [v for v, _ in result["a"]]
    assert pytest.approx(pts[20:81], rtol) == [v for _, v in result["a"]][20:81]


def test_all_steep_curve2():
    # preparation
    prefix = "./outs/test_all_steep_curve2_"
    np.random.seed(42)

    xs = np.linspace(-1, 1, 101)  # sweep-axis, like dc bias
    fs = np.linspace(100, 200, 201)  # frequency-axis
    pts = np.full_like(
        xs,
        200 * xs**2 + np.average(fs),  # 上からきて上に抜ける場合
    )  # dip point: y=y0 (constant)
    rtol = 0.01  # relative tolerance for analysis

    amps = 20 * np.ones_like(xs)
    spec = np.array(
        [a / (0.05**2 + (fs - f0) ** 2) for a, f0 in zip(amps, pts)]
    ).T  # 形状固定のLorentzian
    spec += np.random.normal(size=spec.shape)
    mag = 100 * np.log10(np.abs(spec))

    # run
    apply_nofilter(mag, with_plot=True, filename_plot=prefix + "mag.png")
    mag_filtered = apply_sato_filter(
        mag,
        1,
        False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_filtered.png",
    )
    cont_list = find_contours(
        mag_filtered,
        level=0.03,
        threshold_length=10,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_find_contours.png",
    )
    cont_dict = assign_contours(
        cont_list,
        {"a": (245, 255)},
        with_plot=True,
        show=False,
        filename_plot=prefix + "contours.png",
    )
    region_dict = determine_regions(
        mag,
        cont_dict,
        kernel=5,
        with_plot=True,
        show=False,
        filename_plot=prefix + "regions.png",
    )
    result = determine_peak_positions(
        mag,
        region_dict,
        xaxis=xs,
        yaxis=fs,
        eliminate_nan=False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "peaks.png",
    )

    # check
    assert pytest.approx(xs) == [v for v, _ in result["a"]]
    assert pytest.approx(pts[25:76], rtol) == [v for _, v in result["a"]][25:76]


def test_all_isolated_curve():
    # preparation
    prefix = "./outs/test_all_isolated_curve_"
    np.random.seed(42)

    xs = np.linspace(-1, 1, 101)  # sweep-axis, like dc bias
    fs = np.linspace(100, 200, 201)  # frequency-axis
    pts = np.full_like(
        xs, 40 * xs**2 + np.average(fs) - 30
    )  # dip point: y=y0 (constant)
    rtol = 0.01  # relative tolerance for analysis

    amps = 101 * (1 - np.tanh(xs * 10) ** 4)
    print(amps)
    spec = np.array(
        [a / (0.1**2 + (fs - f0) ** 2) for a, f0 in zip(amps, pts)]
    ).T  # 形状固定のLorentzian
    spec += np.random.normal(size=spec.shape)
    mag = 20 * np.log10(np.abs(spec))

    # run
    apply_nofilter(mag, with_plot=True, filename_plot=prefix + "mag.png")
    mag_filtered = apply_sato_filter(
        mag,
        1,
        False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_filtered.png",
    )
    cont_list = find_contours(
        mag_filtered,
        level=0.05,
        threshold_length=10,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_find_contours.png",
    )
    cont_dict = assign_contours(
        cont_list,
        {"a": (4,)},
        with_plot=True,
        show=False,
        filename_plot=prefix + "contours.png",
    )
    region_dict = determine_regions(
        mag,
        cont_dict,
        kernel=5,
        with_plot=True,
        show=False,
        filename_plot=prefix + "regions.png",
    )
    result = determine_peak_positions(
        mag,
        region_dict,
        xaxis=xs,
        yaxis=fs,
        eliminate_nan=False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "peaks.png",
    )

    # check
    assert pytest.approx(xs) == [v for v, _ in result["a"]]
    assert pytest.approx(pts[35:68], rtol) == [v for _, v in result["a"]][35:68]


def test_all_single_disjointcurve():
    # preparation
    prefix = "./outs/test_all_single_disjointcurve_"
    np.random.seed(42)

    xs = np.linspace(-1, 1, 101)  # sweep-axis, like dc bias
    fs = np.linspace(100, 200, 201)  # frequency-axis
    amps = np.tanh(xs) ** 4 * 20
    pts = np.full_like(
        xs, 40 * xs**2 + np.average(fs) - 30
    )  # dip point: y=y0 (constant)
    rtol = 0.01  # relative tolerance for analysis

    spec = np.array(
        [a / (0.05**2 + (fs - f0) ** 2) for a, f0 in zip(amps, pts)]
    ).T  # 形状固定のLorentzian
    spec += np.random.normal(size=spec.shape)
    mag = 20 * np.log10(np.abs(spec))

    # run
    apply_nofilter(mag, with_plot=True, filename_plot=prefix + "mag.png")
    mag_filtered = apply_sato_filter(
        mag,
        1,
        False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_filtered.png",
    )
    cont_list = find_contours(
        mag_filtered,
        level=0.025,
        threshold_length=100,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_find_contours.png",
    )
    cont_dict = assign_contours(
        cont_list,
        {"left": (257,), "right": (230,)},
        with_plot=True,
        show=False,
        filename_plot=prefix + "contours.png",
    )
    region_dict = determine_regions(
        mag,
        cont_dict,
        kernel=5,
        with_plot=True,
        show=False,
        filename_plot=prefix + "regions.png",
    )
    result = determine_peak_positions(
        mag,
        region_dict,
        xaxis=xs,
        yaxis=fs,
        eliminate_nan=False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "peaks.png",
    )

    # check
    assert pytest.approx(xs) == [v for v, _ in result["left"]]
    assert pytest.approx(xs) == [v for v, _ in result["right"]]
    assert pytest.approx(pts[:35], rtol) == [v for _, v in result["left"]][:35]
    assert pytest.approx(pts[:36], rtol) != [v for _, v in result["left"]][:36]
    assert pytest.approx(pts[59:], rtol) == [v for _, v in result["right"]][59:]
    assert pytest.approx(pts[58:], rtol) != [v for _, v in result["right"]][58:]


def test_all_curves():
    # preparation
    prefix = "./outs/test_all_curves_"
    np.random.seed(42)

    xs = np.linspace(-1, 1, 101)  # sweep-axis, like dc bias
    fs = np.linspace(100, 200, 201)  # frequency-axis
    amps = np.tanh(xs) ** 4 * 20
    pts1 = np.full_like(
        xs, 40 * xs**2 + np.average(fs) - 30
    )  # dip point: y=y0 (constant)
    pts2 = pts1 + 20
    rtol = 0.01  # relative tolerance for analysis

    amps = np.tanh(xs) ** 4 * 30
    spec = (
        np.array([25 / (0.05**2 + (fs - f0) ** 2) for f0 in pts1]).T
        + np.array([a / (0.05**2 + (fs - f0) ** 2) for a, f0 in zip(amps, pts2)]).T
    )
    spec += np.random.normal(size=spec.shape)
    mag = 20 * np.log10(np.abs(spec))

    # run
    apply_nofilter(mag, with_plot=True, filename_plot=prefix + "mag.png")
    mag_filtered = apply_sato_filter(
        mag,
        1,
        False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_filtered.png",
    )
    cont_list = find_contours(
        mag_filtered,
        level=0.025,
        threshold_length=100,
        with_plot=True,
        show=False,
        filename_plot=prefix + "mag_find_contours.png",
    )
    cont_dict = assign_contours(
        cont_list,
        {"connected": (192, 202), "left": (361,), "right": (357,)},
        with_plot=True,
        show=False,
        filename_plot=prefix + "contours.png",
    )
    region_dict = determine_regions(
        mag,
        cont_dict,
        kernel=5,
        with_plot=True,
        show=False,
        filename_plot=prefix + "regions.png",
    )
    result = determine_peak_positions(
        mag,
        region_dict,
        xaxis=xs,
        yaxis=fs,
        eliminate_nan=False,
        with_plot=True,
        show=False,
        filename_plot=prefix + "peaks.png",
    )

    # check
    assert pytest.approx(xs) == [v for v, _ in result["left"]]
    assert pytest.approx(xs) == [v for v, _ in result["right"]]
    assert pytest.approx(xs) == [v for v, _ in result["connected"]]
    _pts1 = [v for _, v in result["connected"]]
    assert pytest.approx(pts1, rtol) == _pts1
    assert pytest.approx(pts2[:38], rtol) == [v for _, v in result["left"]][:38]
    assert pytest.approx(pts2[60:], rtol) == [v for _, v in result["right"]][60:]


# --------------------------------------------#


def test__make_enclosure():
    # 横に貫くポリゴン線が2本 → マージして端を閉じた閉ポリゴンにする
    line1_poly = np.array([(3, 0), (5, 5), (4, 10)])  # (y,x)
    line2_poly = np.array([(6, 10), (6, 5), (4, 0)])

    expected_polygon = [
        np.array(
            [
                (4, 0),
                (3, 0),
                (5, 5),
                (4, 10),
                (6, 10),
                (6, 5),
                (4, 0),
            ]
        )
    ]

    assert np.isclose(
        expected_polygon,
        _make_enclosure(polygons=[line1_poly, line2_poly], h=100, w=10 + 1),
    ).all()

    assert np.isclose(
        expected_polygon,
        _make_enclosure(polygons=[line1_poly, line2_poly[::-1]], h=100, w=10 + 1),
    ).all()


def test__make_enclosure__pass():
    # 初めから閉じているものはそのまま返す
    closed_polygon = np.array([(5, 5), (10, 10), (15, 5), (10, 1), (5, 5)])  # y,x

    assert np.isclose(
        closed_polygon,
        _make_enclosure(polygons=[closed_polygon], h=100, w=20),
    ).all()


def test__make_enclosure__single_open_curve():
    # 初めから空いているものは閉じる
    open_polygon = np.array([(1, 0), (5, 5), (10, 5), (15, 0)])  # y,x
    expect_polygon = np.array([(15, 0), (1, 0), (5, 5), (10, 5), (15, 0)])  # y,x

    polygons = _make_enclosure(polygons=[open_polygon], h=100, w=20)

    for p1, p2 in zip([expect_polygon], polygons):
        assert np.isclose(p1, p2).all()


def test__make_enclosure_disjoint():
    # 横に貫くポリゴン線が2本 → マージして端を閉じた閉ポリゴンにする
    line1_poly = np.array([(30, 0), (50, 50), (40, 100)])  # (y,x)
    line2_poly = np.array([(60, 100), (60, 50), (40, 0)])

    # 初めから閉じているものはそのまま返す
    closed_poly = np.array([(5, 5), (10, 10), (15, 5), (10, 1), (5, 5)])  # y,x

    expected_polygons = [
        closed_poly,
        np.array(
            [
                (40, 0),
                (30, 0),
                (50, 50),
                (40, 100),
                (60, 100),
                (60, 50),
                (40, 0),
            ]
        ),
    ]

    polygons = _make_enclosure(
        polygons=[line1_poly, line2_poly, closed_poly], h=1001, w=101
    )
    for p1, p2 in zip(expected_polygons, polygons):
        assert np.isclose(p1, p2).all()
